import numpy as np
import pandas as pd


import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots
import plotly.figure_factory as ff

import fpdf
from fpdf import FPDF
import time

import shutil
import os
import pkg_resources  

def card_summary(df, show_plot=True):
	df_count = df.nunique(axis=0)
	duration_mean = df['duration'].mean() / 60
	duration_total = df['duration'].sum() / 60
	
	card_dict = {
		'n_samples': df_count['file_path'],
		'n_landscapes': df_count['landscape'],
		'n_environments': df_count['environment'],
		'distinct_days': df_count['dt'],
		'total_time_duration': duration_total,
		'mean_time_duration': duration_mean
	}
	

	fig = make_subplots(rows=2, cols=3, subplot_titles=("Samples", "Landscapes", "Environments", "Distinct Days", "Total Duration", "Mean Duration"), 
						specs=[[{'type': 'indicator'}, {'type': 'indicator'}, {'type': 'indicator'}], [{'type': 'indicator'}, {'type': 'indicator'}, {'type': 'indicator'}]])
	# specs=[[{'type': 'indicator'}, {}], [{‘colspan’: 2}, None]]

	trace0 = go.Indicator(
		mode = "number",
		value = card_dict['n_samples'],
		number = {'prefix': ""},
		delta = {'position': "top", 'reference': 320},
		domain = {'x': [0, 1], 'y': [0, 1]})

	trace1 = go.Indicator(
		mode = "number",
		value = card_dict['n_landscapes'],
		number = {'prefix': ""},
		delta = {'position': "top", 'reference': 320},
		domain = {'x': [0, 1], 'y': [0, 1]})

	trace2 = go.Indicator(
		mode = "number",
		value = card_dict['n_environments'],
		number = {'prefix': ""},
		delta = {'position': "top", 'reference': 320},
		domain = {'x': [0, 1], 'y': [0, 1]})

	trace3 = go.Indicator(
		mode = "number",
		value = card_dict['distinct_days'],
		number = {'suffix': ""},
		delta = {'position': "top", 'reference': 320},
		domain = {'x': [0, 1], 'y': [0, 1]})

	trace4 = go.Indicator(
		mode = "number",
		value = card_dict['total_time_duration'],
		number = {'suffix': " min"},
		delta = {'position': "top", 'reference': 320},
		domain = {'x': [0, 1], 'y': [0, 1]})

	trace5 = go.Indicator(
		mode = "number",
		value = card_dict['mean_time_duration'],
		number = {'suffix': " min"},
		delta = {'position': "top", 'reference': 320},
		domain = {'x': [0, 1], 'y': [0, 1]})

	fig.add_trace(trace0, 1, 1)
	fig.append_trace(trace1, 1, 2)
	fig.append_trace(trace2, 1, 3)
	fig.append_trace(trace3, 2, 1)
	fig.append_trace(trace4, 2, 2)
	fig.append_trace(trace5, 2, 3)

	# fig.update_layout(paper_bgcolor = "lightgray")
	if show_plot:
		fig.show()
	
	return card_dict, fig

#-----------------------------------------------------------------------------------------------------------------------------------

def landscape_environment_heatmap(df, show_plot=True):
	df_group = df.groupby(['landscape', 'environment'], as_index=False)['file_path'].count()
	df_group = df_group.rename(columns={"file_path": "count"})
	
	
	df_group_temp = df_group.pivot(index='landscape', columns='environment', values='count')

	fig = px.imshow(df_group_temp, color_continuous_scale='Viridis', text_auto=True, title='Landscape vs Environment Heatmap')
	fig.update_layout(title_x=0.5)

	if show_plot:
		fig.show()
	
	return df_group, fig

#-----------------------------------------------------------------------------------------------------------------------------------

def plot_landscape_histogram(df, show_plot=True):

	fig = px.histogram(df, x="landscape", color="environment", opacity=0.7, title='Ammount of samples by Landscape')
	fig.update_layout(bargap=0.1, title_x=0.5)

	if show_plot:
		fig.show()
	
	return fig

#-----------------------------------------------------------------------------------------------------------------------------------

def plot_landscape_duration(df, show_plot=True):

	fig = px.box(df, x="landscape", y="duration", title='Duration distribution by Landscape')
	fig.update_layout(title_x=0.5)

	if show_plot:
		fig.show()
	
	return fig

#-----------------------------------------------------------------------------------------------------------------------------------

def plot_landscape_daily_distribution(df, show_plot=True):

	fig = px.histogram(df, x="dt", color="landscape", opacity=0.7, title='Ammount of samples by Day and Landscape')
	fig.update_layout(bargap=0.1, title_x=0.5)

	if show_plot:
		fig.show()
	
	return fig

#-----------------------------------------------------------------------------------------------------------------------------------

def plot_environment_histogram(df, show_plot=True):

	fig = px.histogram(df, x="environment", color="landscape", opacity=0.7, title='Ammount of samples by Environment')
	fig.update_layout(bargap=0.1, title_x=0.5)

	if show_plot:
		fig.show()

	return fig

#-----------------------------------------------------------------------------------------------------------------------------------

def plot_environment_duration(df, show_plot=True):

	fig = px.box(df, x="environment", y="duration", title='Duration distribution by Environment')
	fig.update_layout(title_x=0.5)

	if show_plot:
		fig.show()
	
	return fig

#-----------------------------------------------------------------------------------------------------------------------------------

def plot_environment_daily_distribution(df, show_plot=True):

	fig = px.histogram(df, x="dt", color="environment", opacity=0.7, title='Ammount of samples by Day and Environment')
	fig.update_layout(bargap=0.1, title_x=0.5)

	if show_plot:
		fig.show()
	
	return fig

#-----------------------------------------------------------------------------------------------------------------------------------

def plot_duration_distribution(df, show_plot=True):
	group_labels = ['duration'] # name of the dataset

	fig = ff.create_distplot([df['duration'].values], group_labels)
	fig.update_layout(bargap=0.005, title_text='Duration distribution', title_x=0.5)

	if show_plot:
		fig.show()

	return fig

#-----------------------------------------------------------------------------------------------------------------------------------

class PDF(FPDF):

	def footer(self):
		self.set_y(-15)
		self.set_font('Helvetica', 'I', 8)
		self.set_text_color(128)
		self.cell(0, 10, "Generated with <3 by Maui Software - Page " + str(self.page_no()), 0, 0, 'C')



def create_letterhead(pdf, WIDTH, image):
	pdf.image(image, 0, 0, WIDTH)

def create_title(pdf, title, subtitle=None):
	
	# Add main title
	pdf.set_font('Helvetica', 'b', 20)  
	pdf.ln(100)
	pdf.write(5, title)
	pdf.ln(15)
	
	if subtitle is not None:
		# Add subtitle
		pdf.set_font('Helvetica', 'b', 16)  
		pdf.write(5, subtitle)
		pdf.ln(10)
	
	# Add date of report
	pdf.set_font('Helvetica', '', 14)
	pdf.set_text_color(r=128,g=128,b=128)
	today = time.strftime("%d/%m/%Y")
	pdf.write(4, f'{today}')
	
	# Add line break
	pdf.ln(30)

def write_to_pdf(pdf, words):
	
	# Set text colour, font size, and font type
	pdf.set_text_color(r=0,g=0,b=0)
	pdf.set_font('Helvetica', '', 12)
	
	pdf.write(5, words)

def write_subtitle(pdf, words):
	
	# Set text colour, font size, and font type
	pdf.set_text_color(r=0,g=0,b=0)
	pdf.set_font('Helvetica', 'b', 14)
	
	pdf.write(5, words)

def export_file_names_summary_pdf(df, file_name, analysis_title=None, width=210, hight=297):
	
	if not os.path.exists("images_summary_pdf_temp"):
		os.mkdir("images_summary_pdf_temp")
	
	card_dict, fig = card_summary(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/summary1.png", height=300, width=1200)
	df_group, fig = landscape_environment_heatmap(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/summary2.png")

	fig = plot_landscape_histogram(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/landscape1.png", height=400, width=1200)
	fig = plot_landscape_duration(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/landscape2.png", height=400, width=1200)
	fig = plot_landscape_daily_distribution(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/landscape3.png", height=400, width=1200)

	fig = plot_environment_histogram(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/environment1.png", height=400, width=1200)
	fig = plot_environment_duration(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/environment2.png", height=400, width=1200)
	fig = plot_environment_daily_distribution(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/environment3.png", height=400, width=1200)

	fig = plot_duration_distribution(df, show_plot=False)
	fig.write_image("images_summary_pdf_temp/duration1.png", height=400, width=1200)

	# Global Variables
	TITLE = "Audio Files Exploratory Data Analysis"
	SUBTITLE = analysis_title
	WIDTH = width
	HEIGHT = hight

	# Create PDF
	pdf = PDF() # A4 (210 by 297 mm)


	'''
	First Page of PDF
	'''
	# Add Page
	pdf.add_page()

	letterhead_cover = pkg_resources.resource_filename('maui', 'data/letterhead_cover.png')
	letterhead = pkg_resources.resource_filename('maui', 'data/letterhead.png')

	# with pkg_resources.resource_filename('tempfile', 'data/letterhead_cover.png') as f:
	# 	letterhead_cover = f

	# with pkg_resources.resource_filename('tempfile', 'data/letterhead.png') as f:
	# 	letterhead = f

	# Add lettterhead and title
	create_letterhead(pdf, WIDTH, letterhead_cover)
	create_title(pdf, TITLE, SUBTITLE)


	# Add table
	w = 200
	pdf.image("images_summary_pdf_temp/summary1.png", w=w, x=(WIDTH-w)/2)
	pdf.ln(5)

	intro_text = "This report contains a brief exploratory data analysis comprehending the data obtained by audio file names. The objective is to present an overview of the acoustic landscapes and environments of the recordings, as well as their duration. Further analysis such as false color spectrograms and acoustic indices summarization can be performed with Maui Sotware analysis and visualization tools."
	write_to_pdf(pdf, intro_text)


	pdf.add_page()

	create_letterhead(pdf, WIDTH, letterhead)

	pdf.ln(20)
	write_subtitle(pdf, "1. Landscape Analysis")
	pdf.ln(20)
	pdf.image("images_summary_pdf_temp/landscape1.png", w=w, x=(WIDTH-w)/2)
	pdf.ln(5)
	pdf.image("images_summary_pdf_temp/landscape2.png", w=w, x=(WIDTH-w)/2)
	pdf.ln(5)
	pdf.image("images_summary_pdf_temp/landscape3.png", w=w, x=(WIDTH-w)/2)
	pdf.ln(10)


	pdf.add_page()
	create_letterhead(pdf, WIDTH, letterhead)

	pdf.ln(20)
	write_subtitle(pdf, "2. Environment Analysis")
	pdf.ln(20)
	pdf.image("images_summary_pdf_temp/environment1.png", w=w, x=(WIDTH-w)/2)
	pdf.ln(5)
	pdf.image("images_summary_pdf_temp/environment2.png", w=w, x=(WIDTH-w)/2)
	pdf.ln(5)
	pdf.image("images_summary_pdf_temp/environment3.png", w=w, x=(WIDTH-w)/2)
	pdf.ln(10)

	pdf.add_page()
	create_letterhead(pdf, WIDTH, letterhead)

	pdf.ln(20)
	write_subtitle(pdf, "3. Duration Analysis")
	pdf.ln(20)
	pdf.image("images_summary_pdf_temp/duration1.png", w=w, x=(WIDTH-w)/2)


	pdf.output(file_name, 'F')
	
	shutil.rmtree('images_summary_pdf_temp')
	
	



import audioread

import numpy as np
import pandas as pd
import random


import os
import glob

import time
import datetime

def get_file_structure_leec(filename):

    file_name_segments = filename.split('_')

    try:
        while True:
            file_name_segments.remove('')
    except ValueError:
        pass

    dict_keys = ['landscape', 'channel', 'date', 'time', 'environment']
    dt_timestamp = datetime.datetime.strptime(file_name_segments[2] + ' ' + file_name_segments[3],"%Y%m%d %H%M%S")
    
    audio_dict = {
        'landscape': file_name_segments[0],
        'environment': file_name_segments[4],
        'channel': file_name_segments[1],
        'timestamp_init': dt_timestamp
    }

    return audio_dict

# ------------------------------------------------

def get_audio_info(audio_path, store_duration=0, perc_sample=1):
    file_dict = None

    if os.path.isfile(audio_path):
        basename = os.path.basename(audio_path)
        filename, file_extension = os.path.splitext(basename)

        file_dict = get_file_structure_leec(filename)
        
        file_dict['timestamp_end'] =  None
        file_dict['duration'] =  None
        
        if (store_duration):
            x = audioread.audio_open(audio_file)
            duration = x.duration
            
            file_dict['timestamp_end'] =  file_dict['timestamp_init'] + datetime.timedelta(seconds=duration)
            file_dict['duration'] =  duration
        
        file_dict['file_path'] = audio_file
        
        df = pd.DataFrame(file_dict, index=[0])

    elif os.path.isdir(audio_path):
        file_dict = []
        
        for file_path in glob.glob(audio_path + '/*.wav'):
            if random.uniform(0, 1) < perc_sample:
                basename = os.path.basename(file_path)
                filename, file_extension = os.path.splitext(basename)

                file_dict_temp = get_file_structure_leec(filename)
                
                file_dict_temp['timestamp_end'] =  None
                file_dict_temp['duration'] =  None

                if (store_duration):
                    x = audioread.audio_open(file_path)
                    duration = x.duration

                    file_dict_temp['timestamp_end'] =  file_dict_temp['timestamp_init'] + datetime.timedelta(seconds=duration)
                    file_dict_temp['duration'] =  duration
                
                file_dict_temp['file_path'] = file_path

                file_dict.append(file_dict_temp)
        
        df = pd.DataFrame(file_dict)
        
    else:
        raise Exception("The input must be a file or a directory")

    return df

# ------------------------------------------------

def store_df(df, file_type, base_dir, file_name):
    
    if (file_type == 'csv'):
        full_path = os.path.join(base_dir, file_name+'.csv')
        df.to_csv(full_path)
        
        return
    
    elif (file_type == 'pickle'):
        full_path = os.path.join(base_dir, file_name+'.pkl')
        df.to_pickle(full_path)
        
        return
    


import numpy as np
import pandas as pd
import math

import gc

from tqdm import tqdm
from multiprocessing import Pool
import multiprocessing as mp
import sys

import maad

import maui.io




def get_acoustic_indices(df, args: dict):

    
    indices_list = args['indices_list']
    
    # create temp lit of indices to keep appending for each row
    indices_temp = {}
    
    for ind in indices_list:
        if ind ==  'median_amplitude_envelope':
            indices_temp['M'] = []
        if ind ==  'temporal_entropy':
            indices_temp['Ht'] = []
        if ind ==  'temporal_activity':
            indices_temp['ACTfract'] = []
            indices_temp['ACTcount'] = []
            indices_temp['ACTmean'] = []
        if ind ==  'temporal_events':
            indices_temp['EVNtFract'] = []
            indices_temp['EVNmean'] = []
            indices_temp['EVNcount'] = []
        if ind ==  'frequency_entropy':
            indices_temp['Hf'] = []
            indices_temp['Ht_per_bin'] = []
        if ind ==  'number_of_peaks':
            indices_temp['NBPeaks'] = []
        if ind ==  'spectral_entropy':
            indices_temp['EAS'] = []
            indices_temp['ECU'] = []
            indices_temp['ECV'] = []
            indices_temp['EPS'] = []
            indices_temp['EPS_KURT'] = []
            indices_temp['EPS_SKEW'] = []
        if ind ==  'spectral_activity':
            indices_temp['ACTspfract_per_bin'] = []
            indices_temp['ACTspcount_per_bin'] = []
            indices_temp['ACTspmean_per_bin'] = []
            indices_temp['MeanACTspfract_per_bin'] = []
            indices_temp['MeanACTspcount_per_bin'] = []
        if ind ==  'spectral_events':
            indices_temp['MeanEVNspFract_per_bin'] = []
            indices_temp['EVNspFract_per_bin'] = []
            indices_temp['EVNspMean_per_bin'] = []
            indices_temp['EVNspCount_per_bin'] = []
            indices_temp['EVNsp'] = []
        if ind ==  'spectral_cover':
            indices_temp['LFC'] = []
            indices_temp['MFC'] = []
            indices_temp['HFC'] = []
        if ind ==  'soundscape_index':
            indices_temp['NDSI'] = []
            indices_temp['ratioBA'] = []
            indices_temp['antroPh'] = []
            indices_temp['bioPh'] = []
        if ind ==  'acoustic_diversity_index':
            indices_temp['ADI'] = []
        if ind ==  'acoustic_eveness_index':
            indices_temp['AEI'] = []
        if ind ==  'temporal_leq':
            indices_temp['Leq'] = []
        if ind ==  'spectral_leq':
            indices_temp['Leqf'] = []
            indices_temp['Leqf_per_bin'] = []
        if ind ==  'tfsd':
            indices_temp['tfsd'] = []
        if ind ==  'more_entropy':
            indices_temp['Ht_Havrda'] = []
            indices_temp['Ht_Renyi'] = []
            indices_temp['Ht_pairedShannon'] = []
            indices_temp['Ht_gamma'] = []
            indices_temp['Ht_GiniSimpson'] = []
            indices_temp['Hf_Havrda'] = []
            indices_temp['Hf_Renyi'] = []
            indices_temp['Hf_pairedShannon'] = []
            indices_temp['Hf_gamma'] = []
            indices_temp['Hf_GiniSimpson'] = []
        if ind ==  'acoustic_gradient_index':
            indices_temp['AGI_xx'] = []
            indices_temp['AGI_per_bin'] = []
            indices_temp['AGI_mean'] = []
            indices_temp['AGI_sum'] = []
        if ind ==  'frequency_raoq':
            indices_temp['RAOQ'] = []
        if ind ==  'region_of_interest_index':
            indices_temp['ROItotal'] = []
            indices_temp['ROIcover'] = []
        if ind ==  'acoustic_complexity_index':
            indices_temp['ACI_sum'] = []
            indices_temp['ACI_xx'] = []
            indices_temp['ACI_per_bin'] = []
        

            
    
    # for each row of the dataframe, a list of acoustic indices (given by indices_list parameter) is calculated 
    iterator_df = range(df.shape[0]) if args['parallel'] else tqdm(range(df.shape[0]))

    for index in iterator_df:
#     for index, row in df.iterrows(): 
        row = df.iloc[index]
        
        # load audio file
        s, fs = maad.sound.load(row['file_path'])
        if (len(s) == 0):
            for key in indices_temp:
                indices_temp[key].append(None)
                
        else:
            
        
            # Amplitude Envelope (M)
            if ('median_amplitude_envelope' in indices_list or 'acoustic_richness' in indices_list):

                M_temp = maad.features.temporal_median(s)
                indices_temp['M'].append(M_temp)

            # Temporal Entropy (Ht)
            if ('temporal_entropy' in indices_list or 'acoustic_richness' in indices_list):

                Ht_temp = maad.features.temporal_entropy(s)
                indices_temp['Ht'].append(Ht_temp)


            if ('temporal_leq' in indices_list):

                Leq = maad.features.temporal_leq (s, fs, gain=42)
                indices_temp['Leq'].append(Leq)

            # Temporal Activity (ACT)
            if ('temporal_activity' in indices_list):

                ACTfract, ACTcount, ACTmean = maad.features.temporal_activity(s, 6)


                indices_temp['ACTfract'].append(ACTfract)
                indices_temp['ACTcount'].append(ACTcount)
                indices_temp['ACTmean'].append(ACTmean)

            if ('more_entropy' in indices_list):

                env = maad.sound.envelope(s)
                Ht_Havrda, Ht_Renyi, Ht_pairedShannon, Ht_gamma, Ht_GiniSimpson = maad.features.more_entropy(env**2, order=3)


                indices_temp['Ht_Havrda'].append(Ht_Havrda)
                indices_temp['Ht_Renyi'].append(Ht_Renyi)
                indices_temp['Ht_pairedShannon'].append(Ht_pairedShannon)
                indices_temp['Ht_gamma'].append(Ht_gamma)
                indices_temp['Ht_GiniSimpson'].append(Ht_GiniSimpson)


            # Temporal Events (EVN)
            if ('temporal_events' in indices_list):

                EVNtFract, EVNmean, EVNcount, EVN = maad.features.temporal_events (s, fs, 6)


                indices_temp['EVNtFract'].append(EVNtFract)
                indices_temp['EVNmean'].append(EVNmean)
                indices_temp['EVNcount'].append(EVNcount)

            # Acoustic Complexity Index (ACI)
            if (('acoustic_complexity_index' in indices_list) or ('acoustic_diversity_index' in indices_list) or ('acoustic_eveness_index' in indices_list)):
                Sxx, tn, fn, ext = maad.sound.spectrogram (s, fs, mode='amplitude') 

                if ('acoustic_complexity_index' in indices_list):


                    ACI_xx, ACI_per_bin, ACI_sum  = maad.features.acoustic_complexity_index(Sxx)
                    indices_temp['ACI_sum'].append(ACI_sum)
                    indices_temp['ACI_xx'].append(ACI_xx)
                    indices_temp['ACI_per_bin'].append(ACI_per_bin)

                if ('acoustic_diversity_index' in indices_list):

                    ADI  = maad.features.acoustic_diversity_index(Sxx,fn)        
                    indices_temp['ADI'].append(ADI)


                if ('acoustic_eveness_index' in indices_list):

                    AEI  = maad.features.acoustic_eveness_index(Sxx,fn)        
                    indices_temp['AEI'].append(AEI)



            if (('frequency_entropy' in indices_list) or ('number_of_peaks' in indices_list) or ('spectral_activity' in indices_list) or 
                ('spectral_events' in indices_list) or ('spectral_cover' in indices_list) or ('soundscape_index' in indices_list) or 
                ('spectral_leq' in indices_list)  or ('tfsd' in indices_list)  or ('more_entropy' in indices_list)  or 
                ('acoustic_gradient_index' in indices_list)  or ('frequency_raoq' in indices_list)  or ('region_of_interest_index' in indices_list)  ):

                Sxx_power, tn, fn, ext = maad.sound.spectrogram (s, fs) 

                # Frequency Entropy (Hf)
                if ('frequency_entropy' in indices_list):

                    Hf, Ht_per_bin = maad.features.frequency_entropy(Sxx_power)
                    indices_temp['Hf'].append(Hf)
                    indices_temp['Ht_per_bin'].append(Ht_per_bin)


                # Number of Peaks (NBPeaks)
                if ('number_of_peaks' in indices_list): 

                    NBPeaks = maad.features.number_of_peaks(Sxx_power, fn, slopes=6, min_freq_dist=100) 
                    indices_temp['NBPeaks'].append(NBPeaks)

                if ('tfsd' in indices_list): 

                    tfsd = maad.features.tfsd(Sxx_power,fn, tn) 
                    indices_temp['tfsd'].append(tfsd)

                if ('frequency_raoq' in indices_list): 

                    S_power = maad.sound.avg_power_spectro(Sxx_power) 
                    RAOQ = maad.features.frequency_raoq(S_power, fn)
                    indices_temp['RAOQ'].append(RAOQ)


                if ('soundscape_index' in indices_list): 

                    NDSI, ratioBA, antroPh, bioPh  = maad.features.soundscape_index(Sxx_power,fn) 
                    indices_temp['NDSI'].append(NDSI)
                    indices_temp['ratioBA'].append(ratioBA)
                    indices_temp['antroPh'].append(antroPh)
                    indices_temp['bioPh'].append(bioPh)

                if ('acoustic_gradient_index' in indices_list): 

                    AGI_xx, AGI_per_bin, AGI_mean, AGI_sum = maad.features.acoustic_gradient_index(Sxx_power,tn[1]-tn[0])

                    indices_temp['AGI_xx'].append(AGI_xx)
                    indices_temp['AGI_per_bin'].append(AGI_per_bin)
                    indices_temp['AGI_mean'].append(AGI_mean)
                    indices_temp['AGI_sum'].append(AGI_sum)

                if ('spectral_leq' in indices_list): 

                    Leqf, Leqf_per_bin = maad.features.spectral_leq(Sxx_power, gain=42)
                    indices_temp['Leqf'].append(Leqf)
                    indices_temp['Leqf_per_bin'].append(Leqf_per_bin)

                if ('more_entropy' in indices_list): 

                    S_power = maad.sound.avg_power_spectro(Sxx_power)
                    Hf_Havrda, Hf_Renyi, Hf_pairedShannon, Hf_gamma, Hf_GiniSimpson = maad.features.more_entropy(S_power, order=3)

                    indices_temp['Hf_Havrda'].append(Hf_Havrda)
                    indices_temp['Hf_Renyi'].append(Hf_Renyi)
                    indices_temp['Hf_pairedShannon'].append(Hf_pairedShannon)
                    indices_temp['Hf_gamma'].append(Hf_gamma)
                    indices_temp['Hf_GiniSimpson'].append(Hf_GiniSimpson)

                if ('region_of_interest_index' in indices_list): 

                    Sxx_noNoise= maad.sound.median_equalizer(Sxx_power) 
                    Sxx_dB_noNoise = maad.util.power2dB(Sxx_noNoise)
                    ROItotal, ROIcover = maad.features.region_of_interest_index(Sxx_dB_noNoise, tn, fn)

                    indices_temp['ROItotal'].append(ROItotal)
                    indices_temp['ROIcover'].append(ROIcover)



                if (('spectral_activity' in indices_list) or ('spectral_events' in indices_list) or ('spectral_cover' in indices_list)  ):
    #                 print('entrou')

                    Sxx_noNoise= maad.sound.median_equalizer(Sxx_power, extent=ext) 
                    Sxx_dB_noNoise = maad.util.power2dB(Sxx_noNoise)

                    if ('spectral_activity' in indices_list):


                        ACTspfract_per_bin, ACTspcount_per_bin, ACTspmean_per_bin = maad.features.spectral_activity(Sxx_dB_noNoise)  

                        indices_temp['MeanACTspfract_per_bin'].append(np.mean(ACTspfract_per_bin))
                        indices_temp['MeanACTspcount_per_bin'].append(np.mean(ACTspcount_per_bin))
                        indices_temp['ACTspfract_per_bin'].append(ACTspfract_per_bin)
                        indices_temp['ACTspcount_per_bin'].append(ACTspcount_per_bin)
                        indices_temp['ACTspmean_per_bin'].append(ACTspmean_per_bin)

                    if ('spectral_events' in indices_list):

                        EVNspFract_per_bin, EVNspMean_per_bin, EVNspCount_per_bin, EVNsp = maad.features.spectral_events(Sxx_dB_noNoise, 
                                                                                                                         dt=tn[1]-tn[0], dB_threshold=6, 
                                                                                                                         rejectDuration=0.1, extent=ext)  

                        MeanEVNspFract_per_bin = np.mean(EVNspFract_per_bin)

                        indices_temp['MeanEVNspFract_per_bin'].append(MeanEVNspFract_per_bin)
                        indices_temp['EVNspFract_per_bin'].append(EVNspFract_per_bin)
                        indices_temp['EVNspMean_per_bin'].append(EVNspMean_per_bin)
                        indices_temp['EVNspCount_per_bin'].append(EVNspCount_per_bin)
                        indices_temp['EVNsp'].append(EVNsp)

                    if ('spectral_cover' in indices_list):


                        LFC, MFC, HFC = maad.features.spectral_cover(Sxx_dB_noNoise, fn)  

                        indices_temp['LFC'].append(LFC)
                        indices_temp['MFC'].append(MFC)
                        indices_temp['HFC'].append(HFC)




            # Spectral Entropy (ACI)
            if ('spectral_entropy' in indices_list):

                EAS, ECU, ECV, EPS, EPS_KURT, EPS_SKEW = maad.features.spectral_entropy(Sxx_power, fn) 
                indices_temp['EAS'].append(EAS)
                indices_temp['ECU'].append(ECU)
                indices_temp['ECV'].append(ECV)
                indices_temp['EPS'].append(EPS)
                indices_temp['EPS_KURT'].append(EPS_KURT)
                indices_temp['EPS_SKEW'].append(EPS_SKEW)
        
        del(s)
        del(fs)
        gc.collect()


    
    # insert calculated indices into the original dataframe
    for key in indices_temp:
        df[key] = indices_temp[key]
    
    # Acoustic Richness (AR)
    # This index uses the median of amplitude envelope andteporal entropy to be calculated, and it is calculated out of the main loop
    if (('acoustic_richness' in indices_list)):
        AR_list = maad.features.acoustic_richness_index(df['Ht'], df['M'])
        df['AR'] = AR_list
    
    # print(df.info())
    # print(sys.getsizeof(df)/1024)
    
    
    return df
    
#-----------------------------------------------------------------------------------------------------------------------------------


def calculate_acoustic_indices_par_aux(args):
    return get_acoustic_indices(*args)


#-----------------------------------------------------------------------------------------------------------------------------------


def parallelize_dataframe(df, func, args, num_cores):
    
    # slices = math.floor(round(df.shape[0]/num_cores)/num_cores)*num_cores
    slices = math.floor(round(df.shape[0]*df.shape[1] / 8)/8)
    
    # split the dataframe into num_cores parts
    df_split_temp = np.array_split(df, slices)
    
    df_split = []
    
    for i in range(slices):
        df_split.append([df_split_temp[i], args])
    
    # create a pool of processes
    with mp.Pool(num_cores) as pool:
    
        # apply the function on each part of the dataframe
        result = list(tqdm(pool.imap(func, df_split), total=len(df_split)))

        # close the pool and join the processes
        pool.close()
        pool.join()
        
    
    
    df = pd.concat(result)
    
    return df

#-----------------------------------------------------------------------------------------------------------------------------------



def calculate_acoustic_indices(df_init, indices_list: list, store_df=False, base_dir=None, file_name=None, file_type=None, parallel=True):
    
    # check if the selected indices are available
    available_indices = ['median_amplitude_envelope', 'temporal_entropy', 'acoustic_richness', 'temporal_activity', 'temporal_events', 'acoustic_complexity_index', 
                         'frequency_entropy', 'number_of_peaks', 'spectral_entropy', 'spectral_activity', 'spectral_events', 'spectral_cover', 'soundscape_index',
                         'acoustic_diversity_index', 'acoustic_eveness_index', 'temporal_leq', 'spectral_leq', 'tfsd', 'more_entropy', 'acoustic_gradient_index',
                         'frequency_raoq', 'region_of_interest_index']
    
    diff = set(indices_list).difference(available_indices)
    if not (set(indices_list).issubset(available_indices)): raise Exception (f'''The indices {diff} are not available. The list of available indices is: {available_indices}''')
        
    # check if the dataframe contains the correct columns
    necessary_colums = ['file_path']
    
    if not (set(necessary_colums).issubset(df_init.columns)): raise Exception("Your dataset must contain a column named 'file_path', which shows the path for each audio file.")
    
    args = {'indices_list':indices_list, 'total_iterations': df_init.shape[0], 'parallel':parallel}

    if parallel:
        df_processed = parallelize_dataframe(df_init, calculate_acoustic_indices_par_aux, args, mp.cpu_count())
    else: 
        df_processed = get_acoustic_indices(df_init, args)


    if (store_df):
        maui.io.store_df(df_processed, file_type, base_dir, file_name)
    
    return df_processed


import math
import numpy as np
import pandas as pd
import copy
import warnings
import os

import plotly.express as px
from plotly.subplots import make_subplots
import plotly.graph_objects as go

import maad


def indices_radar_plot(df, indices:list, agg_type:str, group_by:list=None, max_cols:int=3, fig_size:dict=None, show_plot:bool=True):
    """
    Create a radar plot (spider plot) comparing multiple indices from a DataFrame.

    Parameters:
    -----------
    df : pandas DataFrame
        The input DataFrame containing the data.

    indices : list of str
        A list of column names representing the indices to be compared.

    agg_type : str
        The type of aggregation to be applied to the data. Supported options are:
        - 'mean': Calculate the mean value for each index.
        - 'median': Calculate the median value for each index.
        - 'stddev': Calculate the standard deviation for each index.
        - 'var': Calculate the variance for each index.
        - 'max': Calculate the maximum value for each index.
        - 'min': Calculate the minimum value for each index.

    group_by : list of str, optional
        A list of column names by which the data should be grouped. Each unique value
        in these columns will be represented as a separate category in the radar plot.
        Defaults to None, indicating no grouping.

    max_cols : int, optional
        The maximum number of columns for the subplots in the radar plot grid. This
        parameter is only relevant when grouping data. Defaults to 3.

    fig_size : dict, optional
        A dictionary specifying the height and width of the radar plot figure. It
        should have 'height' and 'width' keys. Defaults to None.

    show_plot : bool, optional
        Whether to display the radar plot immediately using Plotly. Defaults to True.

    Returns:
    --------
    fig : plotly.graph_objs.Figure
        The Plotly figure object representing the radar plot.

    Raises:
    -------
    AssertionError
        - If 'agg_type' is not one of the supported aggregation options.
        - If a column specified in 'group_by' is not found in the DataFrame.
        - If 'indices' is empty or contains indices not present in the DataFrame.
        - If 'fig_size' dictionary is missing 'height' or 'width' keys.

    Exception
        - If the maximum number of categories to group by exceeds 2.

    Warning
        - If there are more unique categories in 'group_by' than available colors.

    Example:
    --------
    # Create a radar plot comparing 'index1' and 'index2' aggregated by 'category'
    fig = indices_radar_plot(data_df, ['index1', 'index2'], 'mean', group_by=['category'])
    """
    
    # 0. Initial configuration
    # 0.1. Verify if agg_type is available
    agg_options = ['mean', 'median', 'stddev', 'var', 'max', 'min']
    assert agg_type in agg_options, f"'{agg_type}' is not in {agg_options}"
    
    # 0.2. Verify if group_by column is available
    if group_by is not None:
        for col in group_by:
            assert col in df.columns, f"'{col}' is not in {df.column}"
            
        # 0.2.1. Verify if there is a maximum of two categories to group by
        if len(group_by) > 2:
            raise Exception("Sorry, the maximum categories to group by is 2")
        
    # 0.3. Verify if the select indices have been already calculated
    if indices is None or len(indices) == 0: raise Exception("Sorry, the indices list must be non empty.")
    for index in indices:
        assert index in df.columns, f"'{index}' is not in {df.columns}. Verify if it is correctly spelled and if it have been calculated already."
        
    # 0.4. Verify if fig_size is correctly defined (has two keys, height and width)
    if fig_size is not None:
        if 'height' not in fig_size.keys() or 'width' not in fig_size.keys(): raise Exception("fig_size must contain width and height keys.")
    
    # 0.5 Create a copy of the dataframe
    df_unpivot = copy.deepcopy(df)
    
    #--------------------------------------------

    # 1. Normalize columns to better suit the plot
    for index in indices:
        df_unpivot[index] = (df_unpivot[index]-df_unpivot[index].min())/(df_unpivot[index].max()-df_unpivot[index].min())
        
    #--------------------------------------------
        
    # 2. Unpivot dataframe
    df_unpivot = pd.melt(df_unpivot, id_vars=group_by, value_vars=indices)
    
    #--------------------------------------------
        
    # 3. Aggregate data
    gb_cols = ['variable']
    if group_by is not None:
        for col in group_by: gb_cols.append(col)
    
    if agg_type == 'mean':
        df_unpivot = df_unpivot.groupby(gb_cols).mean().reset_index()
    if agg_type == 'median':
        df_unpivot = df_unpivot.groupby(gb_cols).median().reset_index()
    if agg_type == 'stddev':
        df_unpivot = df_unpivot.groupby(gb_cols).std().reset_index()
    if agg_type == 'var':
        df_unpivot = df_unpivot.groupby(gb_cols).var().reset_index()
    if agg_type == 'max':
        df_unpivot = df_unpivot.groupby(gb_cols).var().reset_index()
    if agg_type == 'min':
        df_unpivot = df_unpivot.groupby(gb_cols).var().reset_index()
    
    df_unpivot = df_unpivot.rename(columns={'variable': 'index'})
    
    #--------------------------------------------
        
    # 4. Plot data
    
    n_cols = 1
    n_rows = 1

    if group_by is not None and len(group_by) > 1:
        n_cols = min(len(list(df[group_by[0]].unique())), max_cols)
        n_rows = math.ceil(len(list(df[group_by[0]].unique())) / max_cols) 
        
        

    fig = make_subplots(
            rows=n_rows, cols=n_cols, 
            specs=[[{'type': 'polar'}]*n_cols]*n_rows,
            subplot_titles=[' ']*n_cols*n_rows
        )
    
    col = 1
    row = 1
    showlegend = True
    
    colors = px.colors.qualitative.Plotly
    
    # se não for agrupar    
    if group_by is None:
            df_tmp = copy.copy(df_unpivot)
            
            r = list(df_tmp["value"])
            r.append(r[0])
            theta = list(df_tmp["index"])
            theta.append(theta[0])
            
            fig.add_trace( go.Scatterpolar(r=r, theta=theta, mode='lines'), 
                          row=1, col=1)
    else:
        if len(group_by) == 1: lables_list = list(df[group_by[0]].unique())
        else: 
            lables_list = list(df[group_by[1]].unique())
        if len(lables_list) > len(colors):  warnings.warn("There are more categories than available color, some categories may use the same color")
            
        for ind, category in enumerate(list(df[group_by[0]].unique())): 


            df_tmp = copy.copy(df_unpivot[df_unpivot[group_by[0]] == category])

            if len(group_by) == 1:
                r = list(df_tmp["value"])
                r.append(r[0])
                theta = list(df_tmp["index"])
                theta.append(theta[0])
                
                fig.add_trace( go.Scatterpolar(name = category, r=r, theta=theta, mode='lines', legendgroup=category, showlegend=showlegend), 
                              row=row, col=col)
                if category in lables_list: lables_list.remove(category)
                if len(lables_list) == 0: showlegend = False
                    
            else:

                for j, filter_col in enumerate(list(df[group_by[1]].unique())):

                    df_tmp_final = copy.copy(df_tmp[df_tmp[group_by[1]] == filter_col])
                    
        
                    r = list(df_tmp_final["value"])
                    if len(r) > 0:
                        r.append(r[0])
                        theta = list(df_tmp_final["index"])
                        theta.append(theta[0])

                        if filter_col not in lables_list: showlegend = False
                        else: 
                            lables_list.remove(filter_col)
                            showlegend = True
                            
                        
                        fig.add_trace( go.Scatterpolar(name = filter_col, r=r, theta=theta, mode='lines', legendgroup=filter_col, 
                                                       showlegend=showlegend, line_color = colors[j % len(colors)]), 
                                      row=row, col=col)   
                        fig.update_polars(radialaxis_showticklabels=False)
                            
                fig.layout.annotations[ind]['text'] = category
                fig.layout.annotations[ind]['yshift'] = 25
                
                if col >= max_cols:
                    col = 1
                    row += 1
                else:
                    col += 1
                    
    fig.update_layout(
        title=f'''Radar Plot - Comparisson between indices''',
        title_x=0.5
    )
    
    fig.layout.autosize = True
    if fig_size is not None:
        fig.update_layout(height=fig_size['height'], width=fig_size['width'])
    fig.update_layout(polar = dict(radialaxis = dict(showticklabels = False)))
    
    if show_plot:
        fig.show()
    return fig
    
# --------------------------------------------------------------------------------------------------------------------------------

def indices_histogram_plot(df, indices:list, group_by:str=None, max_cols:int=3, fig_size:dict=None, show_plot:bool=True):
    """
    Create a histogram plot comparing the distribution of selected indices in a DataFrame.

    Parameters:
    -----------
    df : pandas DataFrame
        The input DataFrame containing the data.

    indices : list of str
        A list of column names representing the indices to be compared.

    group_by : str, optional
        A column name by which the data should be grouped. Each unique value in this
        column will be represented as a separate category in the histogram plot.
        Defaults to None, indicating no grouping.

    max_cols : int, optional
        The maximum number of columns for the subplots in the histogram plot grid.
        Defaults to 3.

    fig_size : dict, optional
        A dictionary specifying the height and width of the histogram plot figure. It
        should have 'height' and 'width' keys. Defaults to None.

    show_plot : bool, optional
        Whether to display the histogram plot immediately using Plotly. Defaults to True.

    Returns:
    --------
    fig : plotly.graph_objs.Figure
        The Plotly figure object representing the histogram plot.

    Raises:
    -------
    AssertionError
        - If 'group_by' is specified, but the column is not found in the DataFrame.
        - If 'indices' is empty or contains indices not present in the DataFrame.
        - If 'fig_size' dictionary is missing 'height' or 'width' keys.

    Exception
        - If 'group_by' is specified, but more than one index is provided for comparison.

    Example:
    --------
    # Create a histogram plot for the distribution of 'index1'
    fig = indices_histogram_plot(data_df, ['index1'])

    # Create a grouped histogram plot for the distribution of 'index2' by 'category'
    fig = indices_histogram_plot(data_df, ['index2'], group_by='category')
    """

    
    # 0. Initial configuration
    # 0.1. Verify if group_by column is available
    if group_by is not None:
        assert group_by in df.columns, f"'{group_by}' is not in {df.column}"
        if len(indices) > 1: raise Exception("Sorry, to group by some category, only one index is supported.")
        
            
        
    # 0.2. Verify if the select indices have been already calculated
    if indices is None or len(indices) == 0: raise Exception("Sorry, the indices list must be non empty.")
    for index in indices:
        assert index in df.columns, f"'{index}' is not in {df.columns}. Verify if it is correctly spelled and if it have been calculated already."
        
    # 0.3. Verify if fig_size is correctly defined (has two keys, height and width)
    if fig_size is not None:
        if 'height' not in fig_size.keys() or 'width' not in fig_size.keys(): raise Exception("fig_size must contain width and height keys.")
    
        
    #--------------------------------------------
    
    # 1. Plot data

    if group_by is None:
        n_cols = min(len(indices), max_cols)
        n_rows = math.ceil(len(indices) / max_cols) 
    else:
        n_cols = min(len(df[group_by].unique()), max_cols)
        n_rows = math.ceil(len(df[group_by].unique()) / max_cols) 

    fig = make_subplots(
            rows=n_rows, cols=n_cols, 
            specs=[[{'type': 'histogram'}]*n_cols]*n_rows,
            subplot_titles=[' ']*n_cols*n_rows
        )
    
    col = 1
    row = 1
    showlegend = True
    
    colors = px.colors.qualitative.Plotly
    if group_by is not None: lables_list = list(df[group_by].unique())
    
    # add one trace for each index
    if group_by is None: 
        for i, index in enumerate(indices):

            fig.add_trace( go.Histogram(name = index, x=df[index], showlegend=showlegend), 
                                         row=row, col=col)   
            fig.layout.annotations[i]['text'] = index
            fig.layout.annotations[i]['yshift'] = 25

            if col >= max_cols:
                col = 1
                row += 1
            else:
                col += 1
    else:
        for i, category in enumerate(df[group_by].unique()):
            df_index = df[df[group_by] == category]

            fig.add_trace( go.Histogram(name = index, x=df_index[indices[0]], showlegend=showlegend), 
                                         row=row, col=col)   
            fig.layout.annotations[i]['text'] = category
            fig.layout.annotations[i]['yshift'] = 25

            if col >= max_cols:
                col = 1
                row += 1
            else:
                col += 1
    
    fig.update_layout(
        title=f'''Histogram Plot - Distribution of selected indices''',
        title_x=0.5
    )
        
        
    fig.layout.autosize = True
    if fig_size is not None:
        fig.update_layout(height=fig_size['height'], width=fig_size['width'])
    fig.update_layout(polar = dict(radialaxis = dict(showticklabels = False)))
    
    if show_plot:
        fig.show()
    return fig
        
# --------------------------------------------------------------------------------------------------------------------------------

def indices_violin_plot(df, indices:list, group_by:str=None, fig_size:dict=None, show_plot:bool=True):
    """
    Create a violin plot comparing the distribution of selected indices in a DataFrame.

    Parameters:
    -----------
    df : pandas DataFrame
        The input DataFrame containing the data.

    indices : list of str
        A list of column names representing the indices to be compared.

    group_by : str, optional
        A column name by which the data should be grouped. Each unique value in this
        column will be represented as a separate category in the violin plot.
        Defaults to None, indicating no grouping.

    fig_size : dict, optional
        A dictionary specifying the height and width of the violin plot figure. It
        should have 'height' and 'width' keys. Defaults to None.

    show_plot : bool, optional
        Whether to display the violin plot immediately using Plotly. Defaults to True.

    Returns:
    --------
    fig : plotly.graph_objs.Figure
        The Plotly figure object representing the violin plot.

    Raises:
    -------
    AssertionError
        - If 'group_by' is specified, but the column is not found in the DataFrame.
        - If 'indices' is empty or contains indices not present in the DataFrame.
        - If 'fig_size' dictionary is missing 'height' or 'width' keys.

    Example:
    --------
    # Create a violin plot for the distribution of 'index1'
    fig = indices_violin_plot(data_df, ['index1'])

    # Create a grouped violin plot for the distribution of 'index2' by 'category'
    fig = indices_violin_plot(data_df, ['index2'], group_by='category')
    """
 
    
    # 0. Initial configuration
    # 0.1. Verify if group_by column is available
    if group_by is not None:
        assert group_by in df.columns, f"'{group_by}' is not in {df.column}"
        

    # 0.2. Verify if the select indices have been already calculated
    if indices is None or len(indices) == 0: raise Exception("Sorry, the indices list must be non empty.")
    for index in indices:
        assert index in df.columns, f"'{index}' is not in {df.columns}. Verify if it is correctly spelled and if it have been calculated already."
        
    # 0.4. Verify if fig_size is correctly defined (has two keys, height and width)
    if fig_size is not None:
        if 'height' not in fig_size.keys() or 'width' not in fig_size.keys(): raise Exception("fig_size must contain width and height keys.")
    
    
    #--------------------------------------------
    # 2. Plot data
    
    n_cols = 1
    n_rows = len(indices)

    fig = make_subplots(
            rows=n_rows, cols=n_cols, 
            specs=[[{'type': 'histogram'}]*n_cols]*n_rows,
            subplot_titles=[' ']*n_cols*n_rows
        )
    
    showlegend = True
    
    violin_width = 0
    if fig_size is not None and group_by is not None:
        violin_width = (fig_size['width'] / (len(list(df[group_by].unique())))) / fig_size['width']
    elif fig_size is not None and group_by is None:
        violin_width = 0.3
        
    colors = px.colors.qualitative.Plotly
    
    if group_by is not None: 
        lables_list = list(df[group_by].unique())
        categories = list(df[group_by].unique())
        
        if len(lables_list) > len(colors):  warnings.warn("There are more categories than available color, some categories may use the same color")
        
    
    # add one trace for each index
    for i, index in enumerate(indices):
        
        if group_by is not None:           
            
            
            for j, lab in enumerate(categories):
        
                fig.add_trace(go.Violin(x=df[group_by][df[group_by] == lab],
                                        y=df[index][df[group_by] == lab],
                                        name=lab,
                                        box_visible=True,
                                        meanline_visible=True,
                                        points='all',
                                        scalemode='width',
                                        width=violin_width,
                                        legendgroup=lab,
                                        showlegend=showlegend,
                                        marker_color=colors[j % len(colors)],
                                        line_color=colors[j % len(colors)]
                                       ), row=i+1, col=1)
            
                if lab in lables_list: lables_list.remove(lab)
                if len(lables_list) == 0: showlegend = False
        else:
            fig.add_trace(go.Violin(y=df[index],
                                    name='',
                                    box_visible=True,
                                    meanline_visible=True,
                                    points='all',
                                    scalemode='width',
                                    width=violin_width,
                                    showlegend=False
                                   ), row=i+1, col=1)
            
        fig.layout.annotations[i]['text'] = index
        fig.layout.annotations[i]['yshift'] = 25
        
    fig.update_layout(
        title=f'''Violin Plot - Distribution of selected indices''',
        title_x=0.5
    )

        
        
    fig.layout.autosize = True
    if fig_size is not None:
        fig.update_layout(height=fig_size['height'], width=fig_size['width'])
        
    if show_plot:
        fig.show()
    return fig
        
# --------------------------------------------------------------------------------------------------------------------------------
    
def spectrogram_plot(file_path:str, mode:str=None, window:str='hann', nperseg:int=1024, noverlap:int=None, verbose:bool=False, fig_size:dict=None, show_plot:bool=True):
    """
    Create and display a spectrogram plot for an audio file using the scikit-maad package.

    Parameters:
    -----------
    file_path : str
        The path to the audio file to be processed.

    mode : str, optional
        The mode for spectrogram computation. Supported options are:
        - 'psd': Power Spectral Density (default).
        - 'amplitude': Amplitude.
        - 'complex': Complex representation.
        Defaults to 'psd'.

    window : str, optional
        The window function to be applied to the signal. Defaults to 'hann'.

    nperseg : int, optional
        The number of data points used in each block for the FFT. Defaults to 1024.

    noverlap : int, optional
        The number of points of overlap between blocks. If None, 50% overlap is used by default.

    verbose : bool, optional
        Whether to display verbose output during spectrogram computation. Defaults to False.

    fig_size : dict, optional
        A dictionary specifying the height and width of the spectrogram plot figure.
        It should have 'height' and 'width' keys. Defaults to None.

    show_plot : bool, optional
        Whether to display the spectrogram plot immediately using Plotly. Defaults to True.

    Returns:
    --------
    fig : plotly.graph_objs.Figure
        The Plotly figure object representing the spectrogram plot.

    Raises:
    -------
    AssertionError
        - If 'mode' is not one of the supported mode options.
        - If 'fig_size' dictionary is missing 'height' or 'width' keys.

    Example:
    --------
    # Create a spectrogram plot for an audio file with the default settings
    fig = spectrogram_plot("audio_file.wav")

    # Create a spectrogram plot with custom parameters
    fig = spectrogram_plot("audio_file.wav", mode='amplitude', window='hamming', nperseg=512, noverlap=256)
    """
    
    # 0. Validations
    # 0.1. available modes 
    mode_options = ['psd', 'mean', 'complex']
    assert mode in mode_options, f"'{mode}' is not in {mode_options}"
    
    # 0.2. Verify if fig_size is correctly defined (has two keys, height and width)
    if fig_size is not None:
        if 'height' not in fig_size.keys() or 'width' not in fig_size.keys(): raise Exception("fig_size must contain width and height keys.")
    
    #--------------------------------------------
    
    # 1. Load file
    s, fs = maad.sound.load(file_path)
    
    # 2. Calculate spectrogram
    Sxx,tn,fn,ext = maad.sound.spectrogram(s, fs, nperseg=nperseg, noverlap=noverlap, verbose=verbose, mode=mode, window=window) 
    
    if mode == 'psd':
        Sxx_disp = maad.util.power2dB(Sxx)
    if mode == 'amplitude':
        Sxx_disp = maad.util.amplitude2dB(Sxx)
    if mode == 'complex':
        Sxx_disp = maad.util.amplitude2dB(Sxx)
    
    fig_size = {'height':500, 'width':1200}

    fig = go.Figure(data=go.Heatmap(
            z=Sxx_disp,
            x=tn,
            y=fn,
            colorscale='gray',
            hoverinfo=None
          ))
    
    fig.update_layout(
        title=f'''Spectrogram generate from the file {os.path.basename(file_path)}''',
        title_x=0.5
    )
    
    if fig_size is not None:
            fig.update_layout(height=fig_size['height'], width=fig_size['width'])
            
    if show_plot:
        fig.show()
    return fig
    
    
    
    